@@ -9,10 +9,12 @@ import requests
 global HEADERS
 global models
 global incantations
+global knowledge_bases
 
 HEADERS = None
 models = {}
 incantations = {}
+knowledge_bases = {}
 
 
 BASE_URL = 'https://58shnvgnlf.execute-api.us-east-1.amazonaws.com/sandbox/'
@@ -47,19 +49,36 @@ def get_incantations():
         incantations[incantation['id']] = inc
 
 
+def get_knowledge_bases():
+    url = BASE_URL + 'knowledge_bases'
+    resp = requests.get(url, headers=HEADERS)
+    resp = _format_http_response(resp)
+    for knowledge_base in resp['body']:
+        kb = KnowledgeBase(
+            title=knowledge_base['title'],
+            kb_id=knowledge_base['id'],
+            supp_id = knowledge_base['supp_id'],
+            description=knowledge_base['description']
+        )
+        knowledge_bases[knowledge_base['id']] = kb
+
+
 def init(api_key: str, verbose: bool = False):
     global HEADERS
     global models
     global incantations
+    global knowledge_bases
     HEADERS = {'x-api-key': api_key}
 
     # get the current state with models and incantations
     get_models()
     get_incantations()
+    get_knowledge_bases()
 
     if verbose:
         print ("All models:", [model.title for model in models.values()])
         print ("All incantations:", [incantation.title for incantation in incantations.values()])
+        print ("All knowledge bases:", [knowledge_base.title for knowledge_base in knowledge_bases.values()])
 
 
 # exceptions
@@ -100,9 +119,11 @@ class Incantation:
         payload = {
             'title': self.title,
             'text': self.text,
-            'supp_id': self.supp_id,
-            'description': self.description
         }
+        if self.supp_id is not None:
+            payload['supp_id'] = self.supp_id
+        if self.description is not None:
+            payload['description'] = self.description
         resp = _format_http_response(requests.post(url, headers=HEADERS, json=payload))
         self.incantation_id = resp['body']['id']
         self.is_deployed = True
@@ -125,10 +146,10 @@ class Incantation:
         resp = _format_http_response(requests.patch(url, headers=HEADERS, json=payload))
         return resp['body']
 
-    def delete(self, incantation_id: str):
-        url = BASE_URL + f'incantations/{incantation_id}'
+    def delete(self):
+        url = BASE_URL + f'incantations/{self.incantation_id}'
         _format_http_response(requests.delete(url, headers=HEADERS))
-        #del incantations[incantation_id] TODO: 'NoneType' object does not support item deletion
+        del incantations[self.incantation_id]
 
 
 class Model:
@@ -178,7 +199,6 @@ class Model:
         url = BASE_URL + f'models/{self.model_id}'
         _format_http_response(requests.delete(url, headers=HEADERS))
         del models[self.model_id]
-        return True
 
     def get_instances(self):
         url = BASE_URL + f'models/{self.model_id}/instances'
@@ -196,19 +216,20 @@ class Model:
         return instances
 
     # convenience method
-    def create_instance(self, ai_name: str = "AI"):
-        instance_obj = ModelInstance(model_id=self.model_id, ai_name=ai_name)
+    def create_instance(self, ai_name: str = "AI", knowledge_bases: list = None):
+        instance_obj = ModelInstance(model_id=self.model_id, ai_name=ai_name, knowledge_bases=knowledge_bases)
         resp_body = instance_obj.create()
         return instance_obj
 
 
 class ModelInstance:
-    def __init__(self, ai_name: str, model_id: str = None, supp_id: str = None, description: str = None, instance_id = None):
+    def __init__(self, ai_name: str, knowledge_bases: list = None, model_id: str = None, supp_id: str = None, description: str = None, instance_id = None):
+        self.ai_name = ai_name
+        self.knowledge_bases = knowledge_bases # list of knowledge base ids
+        self.model_id = model_id
         self.supp_id = supp_id
         self.description = description
         self.instance_id = instance_id
-        self.model_id = model_id
-        self.ai_name = ai_name
         self.chat_history = []
         self.get_chat_history()
         self.is_deployed = self.instance_id is not None
@@ -226,10 +247,13 @@ class ModelInstance:
         if self.is_deployed:
             raise Exception('This model instance has already been deployed: ' + self.instance_id)
         url = BASE_URL + f'models/{self.model_id}/instances'
-        payload = {
-            'supp_id': self.supp_id,
-            'description': self.description
-        }
+        payload = {'ai_name': self.ai_name}
+        if self.knowledge_bases is not None:
+            payload['knowledge_bases'] = self.knowledge_bases
+        if self.supp_id is not None:
+            payload['supp_id'] = self.supp_id
+        if self.description is not None:
+            payload['description'] = self.description
         resp = _format_http_response(requests.post(url, headers=HEADERS, json=payload))
         self.instance_id = resp['body']['id']
         self.is_deployed = True
@@ -241,6 +265,8 @@ class ModelInstance:
             raise Exception('This model instance has not been deployed yet. Please run `model_instance.create()` before running `model_instance.update()')
         url = BASE_URL + f'models/{self.model_id}/instances/{self.instance_id}'
         payload = {}
+        if self.knowledge_bases is not None:
+            payload['knowledge_bases'] = self.knowledge_bases
         if self.supp_id is not None:
             payload['supp_id'] = self.supp_id
         if self.description is not None:
@@ -251,7 +277,6 @@ class ModelInstance:
     def delete(self):
         url = BASE_URL + f'models/{self.model_id}/instances/{self.instance_id}'
         resp = _format_http_response(requests.delete(url, headers=HEADERS))
-        del models[self.instance_id]
         return resp['body']
 
     def respond(self, human_input: list):
@@ -264,6 +289,62 @@ class ModelInstance:
         return resp['body']
 
 
+class KnowledgeBase:
+    def __init__(self, title: str, supp_id: str = None, description: str = None, kb_id: str = None):
+        self.title = title
+        self.supp_id = supp_id
+        self.description = description
+        self.kb_id = kb_id
+        self.is_deployed = self.kb_id is not None
+
+    def create(self):
+        if self.is_deployed:
+            raise Exception('This knowledge base has already been deployed: ' + self.kb_id)
+        url = BASE_URL + 'knowledge_bases'
+        payload = {
+            'title': self.title
+        }
+        if self.supp_id is not None:
+            payload['supp_id'] = self.supp_id
+        if self.description is not None:
+            payload['description'] = self.description
+        resp = _format_http_response(requests.post(url, headers=HEADERS, json=payload))
+        self.kb_id = resp['body']['id']
+        self.is_deployed = True
+        return resp['body']
+
+    def add_file(self, content: str, title: str = None, link_to_source: str = None, supp_id: str = None, description: str = None):
+        kb_file = KnowledgeBaseFile(kb_id=self.kb_id, content=content, title=title, link_to_source=link_to_source, supp_id=supp_id, description=description)
+        resp_body = kb_file.create()
+        return resp_body
+
+
+class KnowledgeBaseFile:
+    def __init__(self, kb_id: str, content: str, title: str = None, link_to_source: str = None, supp_id: str = None, description: str = None):
+        self.kb_id = kb_id
+        self.content = content # content is a string
+        self.title = title
+        self.link_to_source = link_to_source
+        self.supp_id = supp_id
+        self.description = description
+
+    def create(self):
+        url = BASE_URL + f'knowledge_bases/{self.kb_id}/files'
+        payload = {
+            'content': self.content
+        }
+        if self.title is not None:
+            payload['title'] = self.title
+        if self.link_to_source is not None:
+            payload['link_to_source'] = self.link_to_source
+        if self.supp_id is not None:
+            payload['supp_id'] = self.supp_id
+        if self.description is not None:
+            payload['description'] = self.description
+        resp = _format_http_response(requests.post(url, headers=HEADERS, json=payload))
+        return resp['body']
+
+
 # use this to handle different formats of model_spec
 def model_spec_preprocessing(model_spec: dict):
     # if incantation_names is in the model_spec, replace it with incantation_ids
@@ -275,9 +356,10 @@ def model_spec_preprocessing(model_spec: dict):
         model_spec["incantation_ids"] = incantation_ids
         del model_spec["incantation_names"]
 
+    """
     # if temperature is in the model_spec, convert it to a string
     if "temperature" in model_spec:
-        model_spec["temperature"] = str(model_spec["temperature"])
+        model_spec["temperature"] = str(model_spec["temperature"])"""
     
     return model_spec
 
@@ -315,13 +397,13 @@ def update_incantation(incantation_title: str, new_text: str = None, new_title:
     return incantation_obj
 
 # create_model() is a convenience function that creates a Model object and then calls its create() method
-def create_model(model_spec: dict):
+def create_model(model_spec: dict, verbose: bool = False):
     #if model_spec["model_title"] in models:
     #    return get_model(model_spec["model_title"])
 
     model_spec = model_spec_preprocessing(model_spec)
     
-    print ("Model spec:", model_spec)
+    if verbose: print ("Model spec:\n", model_spec)
     model_obj = Model(model_spec=model_spec)
     body_resp = model_obj.create()
     return model_obj
@@ -356,6 +438,30 @@ def update_model(model_title: str, new_model_spec: dict):
     resp_body = model_obj.update()
     return model_obj
 
+# create_knowledge_base() is a convenience function that creates a KnowledgeBase object and then calls its create() method
+def create_knowledge_base(title: str, supp_id: str = None, description: str = None):
+    kb = KnowledgeBase(title, supp_id, description)
+    resp = kb.create()
+    return kb
+
+# get_knowledge_base() is a convenience function that returns a KnowledgeBase object for an existing knowledge base, given its title
+def get_knowledge_base(title: str):
+    # create knowledge base name to id map - knowledge_bases is a dict of knowledge base objects keyed on knowledge_base_id
+    kb_from_title = {}
+    for kb_id in knowledge_bases.keys():
+        kb = knowledge_bases[kb_id] # get the KnowledgeBase object
+        title = kb.title
+        if title not in kb_from_title:
+            kb_from_title[title] = kb
+        else:
+            raise Exception('Duplicate knowledge base title: ' + title)
+    
+    if title in kb_from_title:
+        return kb_from_title[title]
+    else:
+        raise Exception('Knowledge base title not found: ' + title)
+
+
 # list_incantations() is a convenience function that returns a dictionary of all Incantation objects for an account
 def list_incantations(verbose=False):
     if verbose:
@@ -377,10 +483,34 @@ def list_instances(model_title: str, verbose=False):
     model = get_model(model_title)
     instances = model.get_instances()
     if verbose:
+        print ("\nModel instances:")
         for instance_obj in instances.values():
-            print (instance_obj.instance_id, instance_obj.chat_history[-2:])
+            print (f"id: {instance_obj.instance_id}\nai name: {instance_obj.ai_name}\nlast chat: {instance_obj.chat_history[:1]}\n")
     return instances
 
+# list_knowledge_bases() is a convenience function that returns a dictionary of all KnowledgeBase objects for an account
+def list_knowledge_bases(verbose=False):
+    if verbose:
+        print ("\nKnowledge bases:")
+        for kb_obj in knowledge_bases.values():
+            print (f"id: {kb_obj.kb_id}\ntitle: {kb_obj.title}\n")
+    return knowledge_bases
+
+# add a file to a knowledge base
+def add_file_to_kb(kb_title: str, file_path: str, file_type: str = None):
+    kb = get_knowledge_base(kb_title)
+    if file_type is None:
+        file_type = file_path.split(".")[-1]
+    
+    if file_type == "txt":
+        with open(file_path, 'r') as f:
+            # read the file and convert it to a string
+            content = f.read()
+    else:
+        raise Exception("File type not supported: " + file_type)
+    
+    kb.add_file(content=content, title="Title")
+
 
 """
 Functions to expose to the user:
@@ -398,7 +528,9 @@ Functions to expose to the user:
     - create_instance()
 - ModelInstance
     - respond()
+- KnowledgeBase
+    - add_file()
 
-We probably need to make users aware of the Incantation, Model, and ModelInstance classes, but we generally want to discourage them from using them directly.
+We probably need to make users aware of the classes, but we generally want to discourage them from using them directly, because it's easier to just use the convenience functions.
 
 """
\ No newline at end of file
